/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

/*
1. install rxtx in linux: apt-get install librxtx-java
2. start app:  java -Djava.library.path=/usr/lib/jni -jar PC_DIMMER_Server.jar
*/

package pc_dimmer_server;

import gnu.io.*;
import java.io.*;
import java.net.ServerSocket;
import java.net.Socket;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Stream;
import org.json.*;

class DoHourTask extends TimerTask
{
    @Override public void run()
    {
        // Executed every hour
        PC_DIMMER_Server.CheckIfNight();
    }
}

class cConfig{
    String RS232Port="/dev/ttyUSB0";
    int RS232Baudrate=115200;
    boolean CheckIfNight=false;
    int ServerPort=10014;
}

public class PC_DIMMER_Server implements Runnable, SerialPortEventListener {
    static cConfig config;

    static CommPortIdentifier portId;
    static int MaxChan = 8;

    InputStream inputStream;
    static OutputStream outputStream;
    static SerialPort serialPort;
    Thread readThread;
  
    static byte[] ChannelValues = new byte[MaxChan];
    
    static Calendar calendar;
    static SimpleDateFormat sdf;
    static boolean IsNight=false;

    static public class SunRiseSet {
        private double pi = Math.PI;
        private double rad = pi / 180.0;
        private double h = -(50.0 / 60.0) * rad;
        private double breite = 47.5;
        private double b = breite * rad;
        private double laenge = 8.5;
        private int zeitzone = 2;
        private int tag;

        private String aufgangStr;
        private String untergangStr;

        public SunRiseSet(double breite, double laenge, int tag, int zeitzone) {
            this.breite = breite;
            this.laenge = laenge;
            this.zeitzone = zeitzone;
            this.tag = tag;
            init();
        }
        
        private void init() {
            double aufgang = aufgang1(this.tag);
            double untergang = untergang1(this.tag);

            aufgang = aufgang - laenge / 15.0 + zeitzone;
            untergang = untergang - laenge / 15.0 + zeitzone;

            int aufgangh = (int) Math.floor(aufgang);
            String aufgangH = ((aufgangh < 10) ? "0" + String.valueOf(aufgangh) : String.valueOf(aufgangh));
            int aufgangm = (int) Math.floor(((aufgang - aufgangh) * 100) * 3 / 5);
            String aufgangM = ((aufgangm < 10) ? "0" + String.valueOf(aufgangm) : String.valueOf(aufgangm));
            int aufgangs = (int) Math.floor(((((((aufgang - aufgangh) * 100) * 3 / 5) - aufgangm) * 100) * 3 / 5));
            String aufgangS =((aufgangs < 10) ? "0" + String.valueOf(aufgangs) : String.valueOf(aufgangs));
            this.aufgangStr = aufgangH + ":" + aufgangM + ":" + aufgangS;

            int untergangh = (int) Math.floor(untergang);
            String untergangH = ((untergangh < 10) ? "0" + String.valueOf(untergangh) : String.valueOf(untergangh));
            int untergangm = (int) Math.floor(((untergang - untergangh) * 100) * 3 / 5);
            String untergangM = ((untergangm < 10) ? "0" + String.valueOf(untergangm) : String.valueOf(untergangm));
            int untergangs = (int) Math.floor(((((((untergang - untergangh) * 100) * 3 / 5) - untergangm) * 100) * 3 / 5));
            String untergangS = ((untergangs < 10) ? "0" + String.valueOf(untergangs) : String.valueOf(untergangs));
            this.untergangStr = untergangH + ":" + untergangM + ":" + untergangS;

        }

        private double sonnendeklination(double t) {
            return 0.40954 * Math.sin(0.0172 * (t - 79.35));
        }

        private double zeitdifferenz(double deklination) {
            return 12.0 * Math.acos((Math.sin(h) - Math.sin(b) * Math.sin(deklination)) / (Math.cos(b) * Math.cos(deklination))) / pi;
        }

        private double zeitgleichung(double t) {
            return -0.1752 * Math.sin(0.033430 * t + 0.5474) - 0.1340 * Math.sin(0.018234 * t - 0.1939);
        }

        private double aufgang1(double t) {
            double deklination = sonnendeklination(t);
            return 12 - zeitdifferenz(deklination) - zeitgleichung(t);
        }

        private double untergang1(double t) {
            double deklination = sonnendeklination(t);
            return 12 + zeitdifferenz(deklination) - zeitgleichung(t);
        }

        public String getAufgangStr() {
            return aufgangStr;
        }

        public String getUntergangStr() {
            return untergangStr;
        }
    }

    private static int TimeToSeconds(String s) {
        String[] hourMin = s.split(":");
        int hours = Integer.parseInt(hourMin[0]);
        int minutes = Integer.parseInt(hourMin[1]);
        int seconds = Integer.parseInt(hourMin[2]);
        return (hours * 3600 + minutes * 60 + seconds);
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        System.out.println("PC_DIMMER_Server v2.0 vom 28.02.2020");

        // load configuration from XML-file
        System.out.println("Loading configuration file " + System.getProperty("user.dir") + "/settings.json" + "...");
        try{
            JSONObject jsonconfig;
            boolean CreateConfigFile=false;

            File configfile = new File(System.getProperty("user.dir") + "/settings.json");
            if (configfile.exists()){
                // load config-file and parse JSON-content
                StringBuilder contentBuilder = new StringBuilder();
                Stream<String> stream = Files.lines( Paths.get(System.getProperty("user.dir") + "/settings.json"), StandardCharsets.UTF_8);
                stream.forEach(s -> contentBuilder.append(s).append("\n"));
                jsonconfig = new JSONObject(contentBuilder.toString());
                stream.close();
            }else{
                System.out.println("Create new config-file...");
                CreateConfigFile=true;
                jsonconfig = new JSONObject();
            }
            
            // check if all config-entries are available. Otherwise create them
            if (!jsonconfig.has("RS232Port")){jsonconfig.put("RS232Port", "/dev/ttyUSB0");}
            if (!jsonconfig.has("RS232Baudrate")){jsonconfig.put("RS232Baudrate", 115200);}
            if (!jsonconfig.has("CheckIfNight")){jsonconfig.put("CheckIfNight", true);}
            if (!jsonconfig.has("ServerPort")){jsonconfig.put("ServerPort", 10014);}
            
            // copy JSON-config to config-class
            config = new cConfig();
            config.RS232Port=jsonconfig.getString("RS232Port");
            config.RS232Baudrate=jsonconfig.getInt("RS232Baudrate");
            config.CheckIfNight=jsonconfig.getBoolean("CheckIfNight");
            config.ServerPort=jsonconfig.getInt("ServerPort");

            System.out.println("Config: RS232-Port     = " + config.RS232Port);
            System.out.println("Config: RS232-Baudrate = " + Integer.toString(config.RS232Baudrate));
            System.out.println("Config: CheckIfNight   = " + Boolean.toString(config.CheckIfNight));
            System.out.println("Config: Server-Port    = " + Integer.toString(config.ServerPort));
            
            if (CreateConfigFile){
                // write new config to file
                BufferedWriter writer = new BufferedWriter(new FileWriter(System.getProperty("user.dir") + "/settings.json"));
                jsonconfig.write(writer);
                writer.close();
            }
        }catch(JSONException | IOException error){
            System.out.println(error.toString());
        }

        int c;
        for(c=0;c<MaxChan;c++){
            ChannelValues[c]=cbytetojavabyte(0);
        }
        
        // Timer starten
        if (config.CheckIfNight){
            Timer timer_hour = new Timer();
            timer_hour.schedule( new DoHourTask(), /*Delay*/0, /*Interval*/1000*60*60); // Alle 60 Minuten aufrufen
        }
        
        try
        {
            portId = CommPortIdentifier.getPortIdentifier(config.RS232Port);
            PC_DIMMER_Server reader = new PC_DIMMER_Server();
        }catch(Exception e)
        {
            System.out.println("Exception on opening port " + portId);
        }
    }
    
    public PC_DIMMER_Server() {
        try {
            serialPort = (SerialPort) portId.open("PC_DIMMER_Server", 2000);
        } catch (PortInUseException e) {}

        try {
            inputStream = serialPort.getInputStream();
        } catch (IOException e) {}
        try {
            serialPort.addEventListener(this);
        } catch (TooManyListenersException e) {}
        serialPort.notifyOnDataAvailable(true);

        try {
            serialPort.setSerialPortParams(config.RS232Baudrate, SerialPort.DATABITS_8, SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);
        } catch (UnsupportedCommOperationException e){
        }

        System.out.println("Starting main-thread...");
        
        readThread = new Thread(this);
        readThread.start();
    }

    public static byte cbytetojavabyte(int value){
        return (byte)(value & 0xFF);
    }

    public static int javabytetocbyte(byte value){
        return (value & 0xFF);
    }

    static public void Senddata(int Channel, int Startvalue, int Endvalue, long Fadetime){
        float T;
        int Nsteps;
        float Temp;
        int Stp=0;
        long Ftime=0;
        byte[] rs232data = new byte[6];

/*
        // check if the startvalue is different to current value
        if ((Startvalue>-1) && (Startvalue!=ChannelValues[Channel-1])){
            // first set startvalue

            rs232data[0] = cbytetojavabyte(Channel + 127);
            rs232data[1] = cbytetojavabyte(0);
            rs232data[2] = cbytetojavabyte(0);
            rs232data[3] = cbytetojavabyte(0);
            rs232data[4] = cbytetojavabyte(0);
            rs232data[5] = cbytetojavabyte(((int)((255-Startvalue)/2) & 127));

            try{
                outputStream = serialPort.getOutputStream();
                outputStream.write(rs232data);
                outputStream.close();
            }catch(IOException error){
                System.out.println("IO exception in main-thread!");
            }

            // give dimmerpack some time for processing
            try{
                Thread.sleep(500);
            }catch(Exception e){}
        }
*/
        // calculate the fading-steps for dimmerpack                
        T = Fadetime / 10;
        Nsteps = Math.abs(ChannelValues[Channel-1] - Endvalue); // Startvalue-Endvalue

        if((Nsteps != 0) && (T != 0)){
            do{
                Stp=Stp + 1;

                Ftime = Math.round((T * Stp) / Nsteps);
                Temp = Math.abs(((T - ((Ftime * Nsteps) / Stp)) * 100) / T);
            }while(!((Temp <= 15) || (Stp >= 4)));
        }else{
           Ftime = 0;
           Stp=Stp+1;
        }

        // fill the rs232-packet and send it to dimmerpack
        rs232data[0] = cbytetojavabyte(Channel + 127);
        rs232data[1] = cbytetojavabyte(Stp - 1);
        rs232data[2] = cbytetojavabyte((int)(Ftime >>> 14));
        rs232data[3] = cbytetojavabyte((int)(Ftime >>> 7));
        rs232data[4] = cbytetojavabyte((int)(Ftime));
        rs232data[5] = cbytetojavabyte(((int)((255-Endvalue)/2)));

        rs232data[0] |= 0b10000000;
        rs232data[1] &= ~0b10000000;
        rs232data[2] &= ~0b10000000;
        rs232data[3] &= ~0b10000000;
        rs232data[4] &= ~0b10000000;
        rs232data[5] &= ~0b10000000;
        
        try{
            outputStream = serialPort.getOutputStream();
            outputStream.write(rs232data);
            outputStream.close();
        }catch(IOException error){
            System.out.println("IO exception in main-thread!");
        }catch(NullPointerException error) {
        	System.out.println("NullPointerException in main-thread!");
        }
        
        ChannelValues[Channel-1]=cbytetojavabyte(Endvalue);
    }
    
    public void run() {
        while(true){
            ServerSocket CMDSocket=null;
            Socket connectionSocket=null;

            try
            {
                // Socket etablieren
                CMDSocket = new ServerSocket(config.ServerPort); // neuen Server-Socket anlegen
                //CMDSocket.setSoTimeout(10000); // Gibt Timeout, wenn nicht innerhalb der Zeit eine Verbindung hergestellt wird -> macht aber hier keinen Sinn, da wir unendlich auf neue Verbindungen warten wollen

                while(true)
                {
                    try
                    {
                        connectionSocket = CMDSocket.accept(); // Eingehende Verbindungen akzeptieren (Thread wartet hier so lange, bis eine neue Verbindung eingeht!!!)
                        connectionSocket.setSoTimeout(30000); // 30 Sekunden Timeout fürs Datenlesen setzen (GPRS ist evtl. langsam)
                    } catch (IOException e) {}
                    // start new thread for the new client
                    new ClientThread(connectionSocket).start();
                }
            }catch (IOException error){}
        }
    }

    public class ClientThread extends Thread {
        protected Socket connectionSocket;

        public ClientThread(Socket clientSocket) {
            this.connectionSocket = clientSocket;
        }

        @Override
        public void run() {
            int ReceiveLength;
            char[] Buffer;
            String ReceivedData;
            String Answer;
            boolean KeepThreadAlive=true;
            
            try
            {
                // mit offenem Socket verbinden
                BufferedReader inFromClient = new BufferedReader(new InputStreamReader(connectionSocket.getInputStream()));
                BufferedWriter outToClient = new BufferedWriter(new OutputStreamWriter(connectionSocket.getOutputStream()));

                // Dauerschleife für Datenempfang (max. 30 Sekunden auf neue Daten warten...)
                while(KeepThreadAlive)
                {
                    // Aktive Verbindung besteht

                    // Daten empfangen
                    ReceiveLength=inFromClient.read(); // Programm wartet hier, bis neue Daten eingehen (max. 30 Sekunden)

                    if (ReceiveLength>0)
                    {
                        Buffer = new char[ReceiveLength];
                        inFromClient.read(Buffer, 0, ReceiveLength);
                        ReceivedData = String.valueOf(Buffer, 0, ReceiveLength);

                        Answer = ProcessTCPCMD(ReceivedData);

                        outToClient.write(Answer.length());
                        outToClient.flush();
                        Buffer = Answer.toCharArray();
                        outToClient.write(Buffer, 0, Answer.length());
                        outToClient.flush();
                    }else{
                        // Fehler bei Verbindung, da eine Datenlänge von 0 gesendet wurde -> Socket wurde beendet
                        KeepThreadAlive=false; // Thread sauber beenden
                    }
                }
            }catch(IOException error){}
        }
    }    

    String ProcessTCPCMD(String cmd){
        String Answer="";

        if (cmd.isEmpty()){
            Answer="?";
        }else if (cmd.contains(":")){
            // C1:15>64in5000ms;C2:44>64in5000ms;C8:66>32in3000ms;
            String sChannel, sStart, sEnd, sFadetime, cmd_temp;

            // parse all commands for new lamp-values
            cmd_temp=cmd;
            do{
                sChannel=String.copyValueOf(cmd_temp.toCharArray(), cmd_temp.indexOf("C")+1, cmd_temp.indexOf(":")-cmd_temp.indexOf("C")-1);
                sStart=String.copyValueOf(cmd_temp.toCharArray(), cmd_temp.indexOf(":")+1, cmd_temp.indexOf(">")-cmd_temp.indexOf(":")-1);
                sEnd=String.copyValueOf(cmd_temp.toCharArray(), cmd_temp.indexOf(">")+1, cmd_temp.indexOf("in")-cmd_temp.indexOf(">")-1);
                sFadetime=String.copyValueOf(cmd_temp.toCharArray(), cmd_temp.indexOf("in")+2, cmd_temp.indexOf("ms")-(cmd_temp.indexOf("in")+2));

                Senddata(Integer.parseInt(sChannel),Integer.parseInt(sStart),Integer.parseInt(sEnd),Integer.parseInt(sFadetime));
                
                cmd_temp=String.copyValueOf(cmd_temp.toCharArray(), cmd_temp.indexOf(";")+1, cmd_temp.length()-(cmd_temp.indexOf(";")+1));
            }while(cmd_temp.contains(":"));
        }else if (cmd.equals("C?")){
            int i;
            Answer="";
            for(i=0;i<MaxChan;i++){
                Answer=Answer+"C"+String.valueOf(i+1)+"="+String.valueOf(javabytetocbyte(ChannelValues[i]))+";";
            }
        }

        return Answer;
    }    
    
    public void serialEvent(SerialPortEvent event) {
        switch(event.getEventType()) {
            case SerialPortEvent.BI:
            case SerialPortEvent.OE:
            case SerialPortEvent.FE:
            case SerialPortEvent.PE:
            case SerialPortEvent.CD:
            case SerialPortEvent.CTS:
            case SerialPortEvent.DSR:
            case SerialPortEvent.RI:
            case SerialPortEvent.OUTPUT_BUFFER_EMPTY:
                break;
            case SerialPortEvent.DATA_AVAILABLE:
                int c=0;
                int channel=0;
                int endvalue=0;
                
                try {
                    if(inputStream.available()>=6){
                        do{
                            // search for the first data

                            c=inputStream.read();
                            if (c>127){
                                channel=c-127;
                                inputStream.read();
                                inputStream.read();
                                inputStream.read();
                                inputStream.read();
                                endvalue=255-(inputStream.read()*2);

                                if ((channel>0) && (channel<(MaxChan+1))){
                                    ChannelValues[channel-1] = cbytetojavabyte(endvalue);
                                }
                                
                                //System.out.println("Channel " + channel + " is set to " + endvalue);
                            }
                        }while(c>127);
                    }else{
                        // wrong amount of bytes -> flush the stream
                        for(c=0;c<inputStream.available();c++){
                            inputStream.read();
                        }
                    }
                    inputStream.close();
                    
                    } catch (IOException e){
                }
                break;
        }
    }
    
    static void CheckIfNight(){
        // Kalenderfunktionen vorbereiten
        calendar = Calendar.getInstance();
        sdf = new SimpleDateFormat("HH:mm:ss");

        int dayOfYear = calendar.get(Calendar.DAY_OF_YEAR);  
        SunRiseSet sun = new SunRiseSet(50.983, 9.483, dayOfYear, 1);  // Länge, Breite, Tag, Zeitzone
        int SunriseTime = TimeToSeconds(sun.getAufgangStr());
        int CurrentTime = TimeToSeconds(sdf.format(calendar.getTime()));
        int SunsetTime = TimeToSeconds(sun.getUntergangStr());
        
        if (IsNight) {
            // check if it is morning
            if ((CurrentTime>SunriseTime) && (CurrentTime<SunsetTime)){
                IsNight=false;
                Senddata(8, -1, 0, 60000);
            }
        }else{
            // check if it is evening
            if ((CurrentTime>SunsetTime) || (CurrentTime<SunriseTime)){
                IsNight=true;
                Senddata(8, -1, 64, 60000);
            }
        }
    }
}
